use core::net::{IpAddr, Ipv4Addr, Ipv6Addr};
use services::models::Services;
use texting_robots::{get_robots_url, Robot};

const USER_AGENT: &str = "ServiceListerBot";

pub async fn check(url: &str, ipv: Option<u8>) -> Result<Services, String> {
	let local_address = match ipv {
		Some(6) => Some(IpAddr::V6(Ipv6Addr::new(0, 0, 0, 0, 0, 0, 0, 0))),
		Some(4) => Some(IpAddr::V4(Ipv4Addr::new(0, 0, 0, 0))),
		None => None,
		_ => panic!("Wrong IP version"),
	};
	let client = reqwest::Client::builder()
		.http1_only()
		.https_only(true)
		.min_tls_version(reqwest::tls::Version::TLS_1_3)
		.use_rustls_tls()
		.referer(false)
		.local_address(local_address)
		.user_agent(USER_AGENT)
		.timeout(core::time::Duration::new(15, 0))
		.redirect(reqwest::redirect::Policy::none())
		.build()
		.unwrap();

	let robots_url = &get_robots_url(url).unwrap();
	let robots_txt = match get(&client, robots_url).await {
		Ok(response) => response,
		Err(e) => return Err(e.to_string()),
	};

	if robots_txt.status() == 200 {
		let r = Robot::new(USER_AGENT, robots_txt.text().await.unwrap().as_bytes()).unwrap();
		if r.allowed("/") {
			match get(&client, url).await {
				Ok(response) => {
					let canonical_url = response.url();
					let is_root = canonical_url.path() == "/";
					let canonical_url_string = canonical_url.to_string();

					let server = match response.headers().get("server") {
						Some(header) => match header.to_str() {
							Ok(h) => h,
							Err(_) => return Err("Unstringable server header".to_owned()),
						},
						None => "",
					}
					.to_owned();

					let response_body = response.text().await.unwrap();
					Ok(Services {
						url: canonical_url_string,
						server,
						software: if response_body.contains(r#"<meta property="og:platform" content="PeerTube""#) {
							if is_root {
								"peertube"
							} else {
								return Err("Please only use root URL when adding software".to_owned());
							}
						} else if response_body.contains(r#"<input type="text" id="padname" maxlength="50" autofocus x-webkit-speech>"#) {
							"etherpad"
						} else if response_body.contains(r#"<meta name="description" content="An alternative front-end to YouTube">"#) {
							"invidious"
						} else if response_body.contains(r#"<meta name="application-name" content="LibreQR">"#) {
							"libreqr"
						} else if response_body.contains(r#"Based on <a rel="external" href="https://servnest.niv.re/" class="common"><span aria-hidden="true">🌳</span>&nbsp;ServNest</a>"#) {
							"servnest"
						} else if response_body.contains(r#"<link rel="icon" sizes="128x128" href="/img/lufi128.png">"#) {
							"lufi"
						} else if response_body.contains(r#"<link title="LibreX search" type="application/opensearchdescription+xml" href="/opensearch.xml?method=POST" rel="search"/>"#) {
							"librex"
						} else if response_body.contains(r#"<p class="powered-by-forgejo">"#) {
							"forgejo"
						} else if response_body.contains(r#">Powered by Gitea</a>"#) {
							"gitea"
						} else {
							return Err("Unknown software".to_owned());
						}.to_owned(),
						ipv6: String::new(),
						ipv4: String::new(),
						availability_ipv6: String::new(),
						availability_ipv4: String::new(),
						address_ipv6: String::new(),
						address_ipv4: String::new(),
					})
				}
				Err(e) => Err(e.to_string()),
			}
		} else {
			Err("Bot forbidden in /robots.txt".to_owned())
		}
	} else {
		Err("/robots.txt doesn't exist".to_owned())
	}
}

async fn get(client: &reqwest::Client, url: &str) -> Result<reqwest::Response, reqwest::Error> {
	client.get(url.to_owned()).send().await
}
